////////////////////////////////////////////////////////////////////////
/// \file WaveformProcessor_service.cc
///
///  see Calibrator/func/ADCShapeFitTable
///      Calibrator/art/Calibrator_service
///
/// \author ram2aq@virginia.edu
////////////////////////////////////////////////////////////////////////

#include <iostream>

#include "DDTproto/WaveformProcessor.h"

template<class T> inline T sqr(T x){return x*x;}

namespace novaddt
{

  // Number of points we expect to receive, to find the best peak match with
  const unsigned int kNumFineTimingADCPoints = 4; // Including the DCS origin

  const double kZeroOffsetSamples = -1;

  const double kSamplesPerOffset = 1./64;

  const unsigned int kNumTDCPerSample = 32;


  // this method is no longer used, but could come in handy later
  /*TDC OffsetToTDC(uint32_t const& t0, uint64_t const& high_time, 
                  uint8_t const& offset)
  {
    TDC retval;
    retval.val = high_time<<32 | (t0 + (uint64_t)(offset * 
                                 (kNumTDCPerSample * kSamplesPerOffset)));
    retval.frac = offset%2 * 50;
    return retval;
  }*/

  //......................................................................
  /// \brief Helper function for \a ADCShapeFit inner loop.
  ///
  /// Also used independently to calculate the best fit adc peak height (the
  /// return value).
  inline double GetExpectations(double t0,
                                double riseTime, double fallTime,
                                const int16_t* obs,
                                double* exps)
  {
    for(unsigned int t = 0; t < kNumFineTimingADCPoints; ++t){
      // This is the ideal ASIC formula from SimpleReadout
      exps[t] = (t > t0) ? exp(-(t-t0)/fallTime)*(1-exp(-(t-t0)/riseTime)) : 0;
    }
    // We can calculate the best fit baseline and normalization at this offset
    // analytically (just write the chisq expression, and take derivatives with
    // norm or base), like so:
    double U, V, W, X, Z;
    U = V = W = X = Z = 0;
    for(unsigned int i = 0; i < kNumFineTimingADCPoints; ++i){
      U += exps[i];
      V += 1;
      W += obs[i];
      X += exps[i]*exps[i];
      Z += obs[i]*exps[i];
    }
    const double norm = (W*U-V*Z)/(U*U-X*V);
    const double base = (W*X-Z*U)/(X*V-U*U);

    // Correct all the expectations for the best guess baseline and norm
    for(unsigned int n = 0; n < kNumFineTimingADCPoints; ++n){
      exps[n] = norm*exps[n]+base;
    }
    return norm;
  }

  //......................................................................
  uint8_t ADCShapeFit(int16_t adc1, int16_t adc2, int16_t adc3,
                      double riseTime, double fallTime,
                      bool& goodTime)
  {
    // The observations, in chronological order
    const int16_t obs[kNumFineTimingADCPoints] = {0, adc1, adc2, adc3};
    double bestchisq = 1e10; // infinity

    // Sane default
    uint8_t bestoffset = -kZeroOffsetSamples/kSamplesPerOffset;
    goodTime = false;

    // Scan through possible offsets of the peak, looking for the best match
    uint8_t offset = 0;
    do{
      // In TDC
      const double t0 = (double(offset)*kSamplesPerOffset+kZeroOffsetSamples);
      // Expectations
      double exps[kNumFineTimingADCPoints];
      const double norm = GetExpectations(t0, riseTime, fallTime, obs, exps);

      if(norm < 0) continue;
      // Assuming equal, uncorrelated, errors on all points. (Including
      // correlations was not found to improve resolution).

      double chisq = 0;
      for(unsigned int i = 0; i < kNumFineTimingADCPoints; ++i)
        chisq += sqr(obs[i]-exps[i]);
      assert(chisq >= 0);

      if(chisq < bestchisq){
        bestchisq = chisq;
        bestoffset = offset;
        goodTime = true;
      }
    } while(++offset != 0);

    if(bestoffset == 0 || bestoffset == 255){
      /*
      std::cerr << "Best time offset is at edge of range. "
                << "If this happens repeatedly there's a "
                << "timing problem. Details: "
                << adc0 << " " << adc1 << " " << adc2 << std::endl;
      */
      bestoffset = -kZeroOffsetSamples/kSamplesPerOffset;
      goodTime = false;
    }

    return bestoffset;
  }


  //---------------------------------------------------------------
  WaveformProcessor::WaveformProcessor(fhicl::ParameterSet const& pset,
                                   art::ActivityRegistry& reg)
     : fSampleTime  (pset.get< double > ("SampleTime"))
     , fRiseTime    (pset.get< double > ("RiseTime"))
     , fFallTime    (pset.get< double > ("FallTime"))
     , fUseMulti    (pset.get< bool >   ("UseMulti"))
     , fUseMulti4ADC(pset.get< bool >   ("UseMulti4ADC"))
  {
      this->reconfigure(pset);
      reg.sPostBeginRun.watch(this, &WaveformProcessor::postBeginRun);
  }
  //----------------------------------------------------------------
  WaveformProcessor::~WaveformProcessor()
  {
  }

  //----------------------------------------------------------------
  void WaveformProcessor::reconfigure(const fhicl::ParameterSet& pset)
  {
    cet::search_path sp("FW_SEARCH_PATH");
    sp.find_file(pset.get< std::string >("ShapeTableFilename"), fShapeTableFilename);
    std::cout << pset.get< std::string > ("ShapeTableFilename") << std::endl;
    assert(!fShapeTableFilename.empty());
  }

  //---------------------------------------------------------------
  void WaveformProcessor::postBeginRun(art::Run const& run)
  {
    if (fUseMulti && !fTableLoaded) { //load the table if it doesn't already exist
      nTableLoads++;
      std::cout << "Loading table from file for " << nTableLoads << " time." << std::endl;
      TFile f(fShapeTableFilename.c_str());
      assert(!f.IsZombie());

      TTree* tr = (TTree*)f.Get("shape_lookup");
      assert(tr);

      const unsigned int N = tr->GetEntries();
      assert(N == kNumTableEntries);

      uint8_t offset;
      bool good;
      uint16_t adc;
      tr->SetBranchAddress("offset", &offset);
      tr->SetBranchAddress("good", &good);
      if (fUseMulti4ADC) 
        tr->SetBranchAddress("adc", &adc);

      for (unsigned int n=0; n<N; n++){
        tr->GetEntry(n);
        fTable[n] = offset;
        fGood[n] = good;
        if (fUseMulti4ADC)
          fADCTable[n] = adc;
      }

      fTableLoaded = true;
    } // end if (!fTableLoaded)
  }


  void WaveformProcessor::process(uint32_t const& t0, uint64_t const& high_time, 
                           int16_t const& adc1, int16_t const& adc2, 
                           int16_t const& adc3, ADC& adcpeak, 
                           bool& goodTime, TDC& retval)
  {
    uint8_t offset=0;

    if(adc3 >= 0 && adc3 < kADC3Max &&
       adc1 >= kADC1Min && adc1 < adc3 &&
       adc2 >= adc3+kADC2RelMin && adc2 < adc3+kADC2RelMax){

      // And this is the offset it'll be at, based on aforementioned geometry
      const int tableIdx = (kADC2RelMax-kADC2RelMin)* // Wedge thickness
        (-kADC1Min*(adc3+1)+ // Rectangular part
         (adc3*(adc3-1))/2+adc1)+ // Triangular part
        (adc2-adc3-kADC2RelMin); // Distance through wedge

      assert(tableIdx >= 0 && tableIdx < int(kNumTableEntries));
      offset = fTable[tableIdx];
      goodTime = fGood[tableIdx];

    }
    else{
      // It's not in our table, perform the fit
      //offset = ADCShapeFit(adc1, adc2, adc3, fRiseTime, fFallTime, goodTime);

      // for speed, instead just use DCS for the 0.5% of hits that miss
      goodTime = false;
      adcpeak = ADC(adc3);
      retval.val = t0; retval.frac = 0;
      return;
    }

    // We need to extract the best normalization, so repeat the relevant part
    // of the inner loop at that offset.
    // --- For speed the normalization has been saved in a different table,
    // --- the use of which is set as a fhicl parameter.
    if(goodTime){
      //Perform the normalization:
      // The observations, in chronological order
      //const int16_t obs[kNumFineTimingADCPoints] = {0, adc1, adc2, adc3};
      //const double t0 = (double(offset)*kSamplesPerOffset+kZeroOffsetSamples);
      //double junk[kNumFineTimingADCPoints]; // Expectations
      //const double norm = GetExpectations(t0, fRiseTime, fFallTime, obs, junk);

      // The value the peak of the curve would have before scaling by norm
      // Analytic calculation from the functional form in ReadoutSim
      //static const double peak = pow(fRiseTime/(fRiseTime+fFallTime), 
      //  fRiseTime/fFallTime)-pow(fRiseTime/(fRiseTime+fFallTime), 
      //  1+fRiseTime/fFallTime);
      //adcpeak = norm*peak;

      if (fUseMulti4ADC) { //Extract from lookup table for speed:
        // And this is the offset it'll be at, based on aforementioned geometry
        const int tableIdx = (kADC2RelMax-kADC2RelMin)* // Wedge thickness
          (-kADC1Min*(adc3+1)+ // Rectangular part
           (adc3*(adc3-1))/2+adc1)+ // Triangular part
          (adc2-adc3-kADC2RelMin); // Distance through wedge
        adcpeak = fADCTable[tableIdx];
      }
      else //Use DCS:
        adcpeak = ADC(adc3);
    }
    else{
      // Unless the fit failed, in which case we're probably safest just doing
      // this.
      adcpeak = ADC(adc3);
    }

    retval.val = (high_time<<32 | (t0 + (uint64_t)(offset * 
                                 (kNumTDCPerSample * kSamplesPerOffset))));
    retval.frac = (uint8_t)(offset%2 * 50);
  }

  void WaveformProcessor::initmulti() { // initialize TDC and ADC members 
                                     //using multipoint
    //check that multipoint can be used
    const uint32_t version = fNano->getHeader()->getVersion();

    static daqdataformats::NanoSliceVersionConvention conv;

    const unsigned int nSamples = conv.getNSamples(version);
    const unsigned int nPretrig = conv.getNPretriggeredSamples(version);

    if ( (nSamples < 3) || // too few samples
         (nSamples <= nPretrig) || //missing peak
         (version == 0) ) //non-multipoint
           { initdcs(); return; }


    TDC tempTDC;
    ADC tempADC;
    uint32_t t0 = fNano->getTimeStamp();
    uint64_t high_time = (uint64_t) fMicro->getHighWord();
    int16_t adc1 = fNano->getValue(1) - fNano->getValue(0);
    int16_t adc2 = fNano->getValue(2) - fNano->getValue(0);
    int16_t adc3 = fNano->getValue(3) - fNano->getValue(0);
    bool goodTime = true;
    process(t0, high_time, adc1, adc2, adc3, tempADC, goodTime, tempTDC);
    if (!goodTime) // something's gone wrong with the fit, use DCS
      initdcs();
    else {
      fTDC = tempTDC;
      fADC = tempADC;
    }
  }

  void WaveformProcessor::initdcs() { // initialize TDC and ADC members 
                                   //using dcs

    //auto trash = fMicro->getFloatingMicroSlice();
    fTDC.val = ((uint64_t)fMicro->getHighWord() << 32)
                | fNano->getTimeStamp();

    fTDC.frac = 0;
    fADC = ADC(fNano->getPulseHeight());

  }

  DEFINE_ART_SERVICE(WaveformProcessor);

} // end namespace novaddt
